library ac_dart.addon_routebuilder;

import 'acdart_shelf_middleware.dart'
    show
        moustacheFile,
        productHostAuthenticator,
        sessionTokenAuthenticator,
        getSessionContext;
import 'package:shelf/shelf.dart'
    show Handler, Middleware, Response, Request, logRequests;
//import 'package:shelf_route/shelf_route.dart';
import 'package:atlassian_connect_host/atlassian_connect_host.dart'
    show
        AcHostHttpClient,
        AcHostService,
        AcSessionContext,
        ConnectAddon,
        HcInstallationEvent;
import 'dart:async' show Future;
import 'package:logging/logging.dart' show Logger;
import 'package:atlassian_connect_host/host_event_model.dart';
import 'package:path/path.dart' as path show join;
import 'package:atlassian_connect_shelf/src/util.dart'
    show pathRelativeToAppRoot;
//import 'package:shelf_bind/shelf_bind.dart' as b
//    show CustomObjectFactory, RequestBody, handlerAdapter;
import 'package:constrain/constrain.dart';
import 'dart:convert';
import 'package:mojito/mojito.dart';
import 'package:mojito/src/router_impl.dart';
import 'package:http_exception/http_exception.dart';

final Logger _log = new Logger('AcDartAddonRouteBuilder');

//typedef ConnectAddon DescriptorFactory(AddonHost addonHost);
const String _DESCRIPTOR_PATH = '/atlassian-connect.json';

final Validator _validator =
    new Validator(); // TODO: this is not a shared instance!!!

final JsonEncoder _prettyJson = new JsonEncoder.withIndent('  ');

class AcDartRouter extends MojitoRouterBuilder {
  final AddonHost _addonHost;
  final String _assetPath;

  AcDartRouter(Function fallbackHandler, String name, routeable,
      RouterAdapter routerAdapter, this._addonHost, this._assetPath,
      {path, middleware})
      : super(
          fallbackHandler, name, path, routerAdapter, routeable, middleware);

  AcDartRouter.create(AcHostService hostService, this._addonHost,
      {bool logRequests: true, String assetPath,
      Map<Type, CustomObjectFactory> customObjects})
      : this._assetPath = assetPath,
        super.create(
            handlerAdapter: _createHandlerAdapter(hostService, customObjects));

  static AcDartRequestContext _context(AcHostService hostService, request) =>
      new AcDartRequestContext(hostService, request);

  static HandlerAdapter _createHandlerAdapter(
      AcHostService hostService, Map<Type, CustomObjectFactory> customObjects) {
    final Map<Type, CustomObjectFactory> _defaultCustomObjects = {
      AcDartRequestContext: (request) => _context(hostService, request),
      AcHostService: (_) => hostService,
      AcSessionContext: getSessionContext,
      AcHostHttpClient: (req) => _context(hostService, req).httpClient,
      AcHostRegistration: (req) => _context(hostService, req).hostRegistration
    };

    final _customObjects = <Type, CustomObjectFactory>{};
    _customObjects.addAll(_defaultCustomObjects);

    if (customObjects != null) {
      _customObjects.addAll(customObjects);
    }

    final _handlerAdapter = handlerAdapter(customObjects: _customObjects,
        // false till https://bitbucket.org/andersmholmgren/constraint/issue/34/constraint-resolver-does-not-cope-with-all
        validateParameters: false);
    return _handlerAdapter;
  }

  /// adds a route to handle the installation lifecycle event. You can override
  /// the [installationPath] if you desire but it is discouraged
  void addInstallationRoute([String installationPath = '/installed']) {
    post(installationPath, _installationHandler);
  }

  // TODO: consolidate
  void addHcInstallationRoute([String installationPath = '/installed']) {
    post(installationPath, _hcInstallationHandler);
  }

  /// adds a route to serve the `atlassian-connect.json` file from the standard
  /// location. You can provide additional variables that you can use in the
  /// moustache template.
  void addDescriptorTemplateRoute(
      {Map<String, String> additionalVariables: const {}}) {
    get(_DESCRIPTOR_PATH, _descriptorTemplateHander(additionalVariables));
  }

  void addDescriptorRoute(ConnectAddon addon) {
    final violations = _validator.validate(addon);
    if (violations.isNotEmpty) {
      final message = violations.map((v) => v.verboseMessage).join('\n');
      throw new ArgumentError('Invalid Addon\n$message');
    }
    get(_DESCRIPTOR_PATH, () => addon);
  }

//  // TODO: factor this to fit into mojito??
//  void addHipchatOAuth2Provider(path, OAuth2CSRFStateStore stateStore,
//      OAuth2TokenStore tokenStore, UriTemplate completionRedirectUrl,
//      SessionIdentifierExtractor sessionIdExtractor, List<String> scopes,
//      {userGrantPath: '/userGrant', authTokenPath: '/authToken',
//      String callbackUrl, bool storeTokens: true}) {
//    final atp = authTokenPath.toString();
//
//    final cb = callbackUrl != null
//        ? callbackUrl
//        : atp.startsWith('/') ? atp.substring(1) : atp;
//
//    final dancer = new OAuth2ProviderHandlers(clientId, oauthProvider,
//        Uri.parse(cb), stateStore, tokenStore, completionRedirectUrl,
//        sessionIdExtractor, scopes, storeTokens: storeTokens);
//
//    addAll((Router r) => r
//      ..get(userGrantPath, dancer.authorizationRequestHandler())
//      ..get(authTokenPath, dancer.accessTokenRequestHandler()), path: path);
//  }

  Future<Response> _installationHandler(
      @RequestBody() InstallationEvent installationEvent,
      AcHostService hostService) async {
    _log.fine('received installation request '
        '${_prettyJson.convert(installationEvent)}');

    try {
      await hostService.registerHost(installationEvent);

      return new Response.ok("installed");
    } catch (e, stackTrace) {
      _log.warning('Registration failed', e, stackTrace);
      // TODO: more specific code and message. e.g 404, 401, 403
      throw new BadRequestException({}, 'Regstration Failed');
    }
    ;
  }

  Future<Response> _hcInstallationHandler(
          @RequestBody() HcInstallationEvent installationEvent,
          AcHostService hostService) async =>
      await _installationHandler(installationEvent, hostService);

  Handler _descriptorTemplateHander(Map<String, String> additionalVariables) {
    final templateVariables = {
      "addonKey": _addonHost.key,
      "baseUrl": _addonHost.baseUrl.toString()
    }..addAll(additionalVariables);

    return moustacheFile(path.join(_assetPath, 'atlassian-connect.json'),
        templateVariables: templateVariables);
  }

//  MojitoRouterBuilder createChild(String name, path, routeable,
//                                  r.RouterAdapter routerAdapter, Middleware middleware) =>
  @override
  AcDartRouter createChild(String name, path, routeable,
      RouterAdapter routerAdapter, Middleware middleware) {
    // createChild(HandlerAdapter handlerAdapter,

    // TODO this is broken. Need to make use of all the passed in params
    return new AcDartRouter(
        fallbackHandler, name, routeable, routerAdapter, _addonHost, _assetPath,
        path: path, middleware: middleware);
  }
}

class AcDartRequestContext {
  final AcHostService hostService;
  final Request request;

  AcDartRequestContext(this.hostService, this.request);

  AcSessionContext get sessionContext => getSessionContext(request);

  Future<AcHostHttpClient> get httpClient =>
      hostService.httpClientForHostKey(sessionContext.hostKey);

  Future<AcHostRegistration> get hostRegistration => hostService
      .findRegistrationByHostKey(sessionContext.hostKey)
      .then((regOpt) => regOpt.getOrElse(() => null));
}

final String _assetDir = pathRelativeToAppRoot('asset');
